#include "lua_lsup.h"

#define check_term(L) \
    *(LSUP_Term **)luaL_checkudata(L, 1, "LSUP.Term")


LSUP_Term **allocate_term (lua_State *L)
{
    LSUP_Term **tp = lua_newuserdatauv (L, sizeof (*tp), 1);
    luaL_getmetatable (L, "LSUP.Term");
    lua_setmetatable (L, -2);

    return tp;
}


/*
 * Factory methods.
 */

static int l_iriref_new (lua_State *L)
{
    const char *data = luaL_checkstring (L, 1);
    // TODO handle nsm.
    LSUP_NSMap *nsm = lua_touserdata (L, 2);

    LSUP_Term **tp = allocate_term (L);

    *tp = LSUP_iriref_new (data, nsm);
    LUA_NLCHECK (*tp, "Error creating term.");

    return 1;
}


static int l_iriref_new_abs (lua_State *L)
{
    LSUP_Term
        *root = check_term (L),
        *iri = *(LSUP_Term **)luaL_checkudata (L, 2, "LSUP.Term");

    LSUP_Term **tp = allocate_term (L);

    *tp = LSUP_iriref_absolute (root, iri);
    LUA_NLCHECK (*tp, "Error creating term.");

    return 1;
}


static int l_iriref_new_rel (lua_State *L)
{
    LSUP_Term
        *root = check_term (L),
        *iri = *(LSUP_Term **)luaL_checkudata (L, 2, "LSUP.Term");

    LSUP_Term **tp = allocate_term (L);

    *tp = LSUP_iriref_relative (root, iri);
    LUA_NLCHECK (*tp, "Error creating term.");

    return 1;
}


/** @brief create a new literal.
 *
 * Argument 2 (data type) and 3 (lang) are exclusive. If both are specified,
 * datatype has precedence. If both are nil, datatype is set to the default
 * (xsd:string).
 */
static int l_lit_new (lua_State *L)
{
    const char
        *data = luaL_checkstring (L, 1),
        *dtype_str = lua_tostring (L, 2);
    const char *lang = lua_tostring (L, 3);

    LSUP_Term **tp = allocate_term (L);
    LSUP_Term *dtype;

    if (dtype_str) {
        // TODO check memory leak.
        dtype = LSUP_iriref_new (dtype_str, LSUP_default_nsm);
        *tp = LSUP_literal_new (data, dtype);
    }
    else if (lang) *tp = LSUP_lt_literal_new (data, (char *)lang);
    else *tp = LSUP_literal_new (data, NULL);

    LUA_NLCHECK (*tp, "Error creating term.");

    return 1;
}


static int l_bnode_new (lua_State *L)
{
    const char *data = luaL_checkstring (L, 1);

    LSUP_Term **tp = allocate_term (L);

    *tp = LSUP_bnode_new (data);
    LUA_NLCHECK (*tp, "Error creating term.");

    return 1;
}


static int new_copy (lua_State *L)
{
    LSUP_Term *src = check_term (L);

    LSUP_Term **tp = allocate_term (L);

    *tp = LSUP_term_copy (src);
    LUA_NLCHECK (*tp, "Error creating term.");

    return 1;
}


/*
 * Class methods.
 */
static int equals (lua_State *L)
{
    LSUP_Term
        *t1 = check_term (L),
        *t2 = *(LSUP_Term **)luaL_checkudata (L, 2, "LSUP.Term");

    lua_pushboolean (L, LSUP_term_equals (t1, t2));

    return 1;
}


static int term_gc (lua_State *L)
{
    LSUP_Term *t = check_term (L);
    if (t) LSUP_term_free (t);

    return 0;
}


static int to_string (lua_State *L)
{
    LSUP_Term *t = check_term (L);
    char *nt_term = NULL;

    CHECK (nt_codec.encode_term (t, NULL, &nt_term), fail);
    lua_pushfstring (L, "LSUP.Term @ %p %s", t, nt_term);

    return 1;

fail:
    if (nt_term) free (nt_term);
    return luaL_error (L, "Error encoding term for display!");
}


static int to_n3 (lua_State *L)
{
    LSUP_Term *t = check_term (L);
    char *nt_term = NULL;

    CHECK (nt_codec.encode_term (t, NULL, &nt_term), fail);
    lua_pushstring (L, nt_term);

    return 1;

fail:
    if (nt_term) free (nt_term);
    return luaL_error (L, "Error encoding term for display!");
}


/*
 * Getters.
 */

// Forward declaration.
static const luaL_Reg term_getters [];


static int get_attr (lua_State *L)
{
    const char *attr = luaL_checkstring (L, 2);
    //printf ("Got attr: %s\n", attr);

    if (luaL_getmetafield (L, 1, attr) != LUA_TNIL)
        return 1;  // Return metamethod to be called as a function.

    luaL_getmetafield (L, 1, "getters");
    if (lua_getfield (L, -1, attr) != LUA_TNIL)
        return lua_tocfunction (L, -1)(L);

    return 0;
}


static int get_data (lua_State *L)
{
    LSUP_Term *t = check_term (L);
    lua_pushstring (L, t->data);

    return 1;
}


static int get_type (lua_State *L)
{
    LSUP_Term *t = check_term (L);
    lua_pushinteger (L, t->type);

    return 1;
}


static int get_iriref_nsm (lua_State *L)
{
    LSUP_Term *t = check_term (L);
    // TODO
    lua_pushlightuserdata (L, LSUP_iriref_nsm (t));

    return 1;
}


static int get_iriref_prefix (lua_State *L)
{
    LSUP_Term *t = check_term (L);
    lua_pushstring (L, LSUP_iriref_prefix (t));

    return 1;
}


static int get_iriref_path (lua_State *L)
{
    LSUP_Term *t = check_term (L);
    lua_pushstring (L, LSUP_iriref_path (t));

    return 1;
}


static int get_iriref_frag (lua_State *L)
{
    LSUP_Term *t = check_term (L);
    lua_pushstring (L, LSUP_iriref_frag (t));

    return 1;
}


static int get_lit_datatype (lua_State *L)
{
    printf ("Getting datatype.\n");
    LSUP_Term *t = check_term (L);
    if (!LSUP_IS_LITERAL (t))
        return luaL_error (L, "Term is not a literal.", 2);

    lua_pushstring (L, t->datatype->data);

    return 1;
}


static int get_lit_lang (lua_State *L)
{
    LSUP_Term *t = check_term (L);
    if (!LSUP_IS_LITERAL (t))
        return luaL_error (L, "Term is not a literal.", 2);

    lua_pushstring (L, t->lang);

    return 1;
}


static int get_hash (lua_State *L)
{
    LSUP_Term *t = check_term (L);

    lua_pushinteger (L, LSUP_term_hash (t));

    return 1;
}


/*
 * Setters.
 */

// Very simple for now.
//static const luaL_Reg term_setters [];
static int set_attr (lua_State *L)
{ return luaL_error (L, "Direct setting is not allowed for this type.", 2); }


/*
 * Library setup.
 */

static const luaL_Reg term_lib_fn [] = {
    {"iriref_new", l_iriref_new},
    {"iriref_new_abs", l_iriref_new_abs},
    {"iriref_new_rel", l_iriref_new_rel},
    {"lit_new", l_lit_new},
    {"bnode_new", l_bnode_new},
    {"term_new_copy", new_copy},

    {NULL}
};


static const luaL_Reg term_getters [] = {
    // General getters.
    {"data", get_data},
    {"type", get_type},

    // IRIRef getters.
    {"nsm", get_iriref_nsm},
    {"prefix", get_iriref_prefix},
    {"path", get_iriref_path},
    {"frag", get_iriref_frag},

    // Literal getters.
    {"datatype", get_lit_datatype},
    {"lang", get_lit_lang},

    // Generic getters.
    {"hash", get_hash},

    {NULL}
};


/*
static const luaL_Reg term_setters [] = {
    {NULL}
};
*/

static const luaL_Reg term_lib_meth [] = {
    {"__eq", equals},
    {"__gc", term_gc},
    {"__index", get_attr},
    {"__tostring", to_string},
    {"__newindex", set_attr},

    {"to_n3", to_n3},
    {NULL}
};


int luaopen_lsup_term (lua_State *L)
{
    LSUP_init();  // This is idempotent: no problem calling it multiple times.
    luaL_newmetatable (L, "LSUP.Term");
    luaL_setfuncs (L, term_lib_meth, 0);

    // Getters table.
    lua_newtable (L);
    for (int i = 0; term_getters[i].name != NULL; i++) {
        lua_pushcfunction (L, term_getters[i].func);
        lua_setfield (L, -2, term_getters[i].name);
    }
    // Set getters table as a value for the Term metatable.
    lua_setfield (L, -2, "getters");

    luaL_newlib (L, term_lib_fn);

    return 1;
}
